# Written by Bram Cohen
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

from time import clock
from binascii import b2a_hex
from BitTornado.bencode import bencode, bdecode
from BitTornado.CurrentRateMeasure import Measure

from BitTornado.bitfield import Bitfield
from BitTornado.__init__ import USERAGENT

DEBUG = False

def toint(s):
    return long(b2a_hex(s), 16)

def tobinary(i):
    return (chr(i >> 24) + chr((i >> 16) & 0xFF) + 
            chr((i >> 8) & 0xFF) + chr(i & 0xFF))

CHOKE = chr(0)
UNCHOKE = chr(1)
INTERESTED = chr(2)
NOT_INTERESTED = chr(3)
# index
HAVE = chr(4)
# index, bitfield
BITFIELD = chr(5)
# index, begin, length
REQUEST = chr(6)
# index, begin, piece
PIECE = chr(7)
# index, begin, length
CANCEL = chr(8)
# 2-byte port message
PORT = chr(9)
# id, dict
EXTENDED = chr(20)

UT_METADATA = 3
METADATAREQUEST = 0
METADATADATA = 1
METADATAREJECT = 2

UT_PEX = 2


class Connection:
    def __init__(self, connection, connecter):
        self.connection = connection
        self.connecter = connecter
        self.got_anything = False
        self.next_upload = None
        self.outqueue = []
        self.partial_message = None
        self.download = None
        self.upload = None
        self.send_choke_queued = False
        self.just_unchoked = None
        self.dht_port = None
        self.ut_metadata = 0
        self.ut_pex = 0
        self.lastpex = 0
        self.prevconnectiondns = []
        # 60 s till first PEX message really sent,
        # then abc.conf utpexsendingrate parameter value
        self.utpexsendingrate = 60

    def get_ip(self, real=False):
        return self.connection.get_ip(real)

    def get_dns(self, real=False):
        return self.connection.get_dns(real)

    def get_id(self):
        return self.connection.get_id()

    def get_readable_id(self):
        return self.connection.get_readable_id()

    def close(self):
        if DEBUG:
            print 'connection closed'
        self.connection.close()

    def is_locally_initiated(self):
        return self.connection.is_locally_initiated()

    def is_encrypted(self):
        return self.connection.is_encrypted()

    def send_interested(self):
        self._send_message(INTERESTED)

    def send_not_interested(self):
        self._send_message(NOT_INTERESTED)

    def send_choke(self):
        if self.partial_message:
            self.send_choke_queued = True
        else:
            self._send_message(CHOKE)
            self.upload.choke_sent()
            self.just_unchoked = 0

    def send_unchoke(self):
        if self.send_choke_queued:
            self.send_choke_queued = False
            if DEBUG:
                print 'CHOKE SUPPRESSED'
        else:
            self._send_message(UNCHOKE)
            if (self.partial_message or self.just_unchoked is None
                or not self.upload.interested or self.download.active_requests):
                self.just_unchoked = 0
            else:
                self.just_unchoked = clock()

    def send_port(self, port):
        self._send_message(PORT + chr((port & 0xFF00) >> 8) + chr(port & 0xFF))

    def send_extended(self):
        extendeddict = {'m': {}, 'p': self.connecter.port, 'v': USERAGENT[0].encode('utf_8'),
                        'e': self.connecter.config['crypto_allowed']}
        if not self.connecter.private:
            extendeddict['m']['ut_metadata'] = UT_METADATA
            if not self.connecter.config['magnet']:
                extendeddict['metadata_size'] = self.connecter.metadatasize
            if self.connecter.config['ut_pex']:
                extendeddict['m']['ut_pex'] = UT_PEX
        #print 'send_extended', extendeddict
        self._send_message(EXTENDED + chr(0) + bencode(extendeddict))
        if self.connecter.config['magnet']:
            self.connection.encoder.startMetadataTimeOut(-1)

    def send_metadatarequest(self, piece):
        #print 'send_metadatarequest', piece
        self._send_message(EXTENDED + chr(self.ut_metadata)
                           + bencode({'msg_type': METADATAREQUEST, 'piece': piece}))
        self.connection.encoder.startMetadataTimeOut(piece)

    def send_metadatadata(self, piece):
        if self.ut_metadata:
            #print 'send_metadatadata', piece
            msg = EXTENDED + chr(self.ut_metadata) \
                  + bencode({'msg_type': METADATADATA, 'piece': piece, 'total_size': self.connecter.metadatasize}) \
                  + self.connecter.metadata[piece * 16384:(piece + 1) * 16384]
            self._send_message(msg)
            self.upload.measure.update_rate(len(msg))
            self.connecter.ratelimiter.measure.update_rate(len(msg))

    def send_metadatareject(self, piece):
        if self.ut_metadata:
            #print 'send_metadatareject', piece
            self._send_message(EXTENDED + chr(self.ut_metadata)
                               + bencode({'msg_type': METADATAREJECT, 'piece': piece}))

    def send_request(self, index, begin, length):
        self._send_message(REQUEST + tobinary(index) + 
                           tobinary(begin) + tobinary(length))
        if DEBUG:
            print 'sent request: ' + str(index) + ': ' + str(begin) + '-' + str(begin + length)

    def send_cancel(self, index, begin, length):
        self._send_message(CANCEL + tobinary(index) + 
                           tobinary(begin) + tobinary(length))
        if DEBUG:
            print 'sent cancel: ' + str(index) + ': ' + str(begin) + '-' + str(begin + length)

    def send_bitfield(self, bitfield):
        self._send_message(BITFIELD + bitfield)

    def send_have(self, index):
        self._send_message(HAVE + tobinary(index))

    def send_keepalive(self):
        self._send_message('')

    def _send_message(self, s):
        #print 'sending', s
        s = tobinary(len(s)) + s
        if self.partial_message:
            self.outqueue.append(s)
        else:
            self.connection.send_message_raw(s)
            #if self.upload:
            #    self.upload.measure.update_rate(len(s))
            #self.connecter.ratelimiter.measure.update_rate(len(s))

    def send_partial(self, bytes):
        if self.connection.closed:
            return 0
        if self.partial_message is None:
            s = self.upload.get_upload_chunk()
            if s is None:
                return 0
            index, begin, piece = s
            self.partial_message = ''.join((tobinary(len(piece) + 9), PIECE,
                                            tobinary(index), tobinary(begin), piece.tostring()))
            self.partial_message_index = 0
            self.partial_message_length = len(self.partial_message)
            if DEBUG:
                print 'sending chunk: ' + str(index) + ': ' + str(begin) + '-' + str(begin + len(piece))

        if self.partial_message_length - self.partial_message_index > bytes:
            self.connection.send_message_raw(self.partial_message[self.partial_message_index:self.partial_message_index + bytes])
            self.partial_message_index += bytes
            return bytes

        if self.send_choke_queued:
            self.send_choke_queued = False
            self.outqueue.append(tobinary(1) + CHOKE)
            self.upload.choke_sent()
            self.just_unchoked = 0

        self.connection.send_message_raw(self.partial_message[self.partial_message_index:] + ''.join(self.outqueue))
        self.outqueue = []
        self.partial_message = None
        return self.partial_message_length - self.partial_message_index

    def get_upload(self):
        return self.upload

    def get_download(self):
        return self.download

    def set_download(self, download):
        self.download = download

    def backlogged(self):
        return not self.connection.is_flushed()

    def got_request(self, i, b, l):
        self.upload.got_request(i, b, l)
        if self.just_unchoked:
            self.connecter.ratelimiter.ping(clock() - self.just_unchoked)
            self.just_unchoked = 0

    def got_extended(self, extendeddict):
        if self.connecter.config['magnet']:
            self.connection.encoder.resetMetadataTimeOut()

        #print 'got extended'
        ut_metadata = extendeddict['m'].get('ut_metadata')
        old_ut_metadata = self.ut_metadata
        if ut_metadata is not None and type(ut_metadata) == int and ut_metadata:
            #print '...supports ut_metadata', ut_metadata
            self.ut_metadata = ut_metadata
            if self.ut_metadata and self.connecter.config['magnet']:
                metadata_size = extendeddict.get('metadata_size')
                if metadata_size is not None:
                    #print '...total size', metadata_size
                    self.connecter.metadatasize = metadata_size
        else:
            #print "...doesn't support ut_metadata"
            self.ut_metadata = 0
        ut_pex = extendeddict['m'].get('ut_pex')
        if ut_pex is not None and type(ut_pex) == int:
            if ut_pex and not self.ut_pex:
                self.prevconnectiondns[:] = []
            self.ut_pex = ut_pex
        cryptoallowed = extendeddict.get('e')
        if cryptoallowed == 0:
            self.cryptoallowed = 0
        elif cryptoallowed == 1:
            self.cryptoallowed = 1
        port = extendeddict.get('p')
        if port is not None and type(port) == int and port != self.connection.remoteport:
            #print '...port', port
            self.connection.remoteport = port

        if self.connecter.config['magnet']:
            if self.ut_metadata:
                if old_ut_metadata == 0:
                    if self.connecter.curmetadatapiece == -1:
                        self.connecter.curmetadatapiece = 0
                    self.send_metadatarequest(self.connecter.curmetadatapiece)
            else:
                self.connection.encoder.storePeerWithNoMetadataSupport()
                self.connection.close()

    def got_metadatarequest(self, piece):
        #print 'got_metadatarequest', piece
        if self.connecter.metadatarequestmeasure.update_rate(1) > 30 \
           or self.connecter.config['magnet'] or piece * 16384 >= self.connecter.metadatasize:
            self.send_metadatareject(piece)
        else:
            self.send_metadatadata(piece)

    def got_metadatadata(self, piece, totalsize, data):
        if self.connecter.curmetadatapiece == -1 or piece != self.connecter.curmetadatapiece \
           or piece != 0 and totalsize != self.connecter.metadatasize:
            #print 'got_metadatadata with bad info'
            self.connection.close()
            return

        self.connection.encoder.resetMetadataTimeOut()

        #print 'got_metadatadata', piece

        if piece == 0:
            self.connecter.metadatasize = totalsize

        self.connecter.metadata.append(data)

        piece += 1
        if piece * 16384 < totalsize:
            if self.ut_metadata:
                self.connecter.curmetadatapiece = piece
                self.send_metadatarequest(piece)
            else:
                self.connection.close()
        else:
            # Metadata complete
            self.connection.close(trynext = False)
            self.connecter.curmetadatapiece = -1
            self.connecter.metadata = ''.join(self.connecter.metadata)
            self.connecter.metadatacompletefunc(self.connecter.metadata)

    def got_metadatareject(self, piece):
        #print 'got_metadatareject', piece
        if piece == self.connecter.curmetadatapiece:
            self.connection.close()

    def got_pex(self, added, addedflags):
        #print 'got_pex'
        if self.connecter.paused:
            return
        oldlastpex = self.lastpex
        self.lastpex = clock()
        if oldlastpex and self.lastpex - oldlastpex < 30:
            self.connection.close()
        peers = []
        for peerstart in xrange(0, len(added), 6):
            ip = '.'.join([str(ord(i)) for i in added[peerstart:peerstart + 4]])
            if ip != '127.0.0.1':
                port = (ord(added[peerstart + 4]) << 8) | ord(added[peerstart + 5])
                peers.append(((ip, port), 0, None, ord(addedflags[peerstart / 6]) & 1))
        #print "...Adding peers", peers
        self.connection.encoder.start_connections(peers)


class Connecter:
    def __init__(self, make_upload, downloader, choker, numpieces,
                 ratelimiter, dht, private, port, metadata,
                 metadatacompletefunc, readbufferallowedfunc, sched, config):
        self.downloader = downloader
        self.make_upload = make_upload
        self.choker = choker
        self.numpieces = numpieces
        self.config = config
        self.ratelimiter = ratelimiter
        self.dht = dht
        self.utpexsendingrate = self.config['ut_pex_sending_rate']
        self.connections = {}
        self.external_connection_made = 0
        self.reported_port = self.config['dhtport']
        self.private = private
        self.port = port
        self.metadata = metadata
        self.metadatasize = len(metadata)
        self.curmetadatapiece = -1
        self.metadatacompletefunc = metadatacompletefunc
        self.readbufferallowedfunc = readbufferallowedfunc
        self.sched = sched
        self.paused = False
        self.metadatarequestmeasure = Measure(60)
        self.pexupmeasure = Measure(self.config['max_rate_period'])
        self.peers = []

    def updateData(self, numpieces):
        self.numpieces = numpieces
        self.metadatasize = len(self.metadata)

    def resetMetadata(self):
        self.metadata = []
        self.metadatasize = 0

    def storePeers(self):
        self.peers[:] = [(con.get_ip(), con.remoteport) for con in self.connections if con.remoteport is not None]

    def how_many_connections(self):
        return len(self.connections)

    def connection_made(self, connection):
        if connection.locally_initiated:
            connection.remoteport = connection.get_port()
        c = Connection(connection, self)
        self.connections[connection] = c
        c.upload = self.make_upload(c, self.ratelimiter, self.readbufferallowedfunc)
        c.download = self.downloader.make_download(c)
        self.choker.connection_made(c)
        if not self.private:
            # Start PEX
            self.sched(self.send_pex, self.utpexsendingrate, [connection])
        if connection.uses_extended:
            #print 'connection supports extended'
            c.send_extended()
        else:
            #print "connection doesn't support extended"
            if self.config['magnet']:
                connection.encoder.storePeerWithNoMetadataSupport()
                connection.close()
        if connection.uses_dht and self.dht:
            c.send_port(self.reported_port)
        self.sched(self.expireUploadPieceBuffer, 30, [connection])
        return c

    def connection_lost(self, connection, shutdown = False):
        c = self.connections[connection]
        del self.connections[connection]
        if c.download:
            c.download.disconnected(shutdown)
        if c.upload:
            c.upload.disconnected()
        self.choker.connection_lost(c, shutdown)
        c.download = None
        c.upload = None
        if shutdown:
            c.next_upload = None

    def connection_flushed(self, connection):
        conn = self.connections[connection]
        if conn.next_upload is None and (conn.partial_message is not None
           or len(conn.upload.buffer) > 0):
            self.ratelimiter.queue(conn)

    def expireUploadPieceBuffer(self, connection):
        c = self.connections.get(connection)
        if c is not None:
            c.upload.expirePieceBuffer()
            self.sched(self.expireUploadPieceBuffer, 30, [connection])

    def setPex(self, flag):
        self.config['ut_pex'] = int(flag)
        if not self.paused:
            self.switchPex(flag)

    def switchPex(self, flag):
        if not self.private:
            if flag:
                flag = UT_PEX
            else:
                flag = 0
            for connection in self.connections:
                if connection.uses_extended:
                    self.connections[connection]._send_message(EXTENDED + chr(0) + bencode({'m': {'ut_pex': flag}}))

    def compactPeerInfoFromDNS(sel, dns):
        return ''.join([chr(int(i)) for i in dns[0].split('.')]) + chr((dns[1] & 0xFF00) >> 8) + chr(dns[1] & 0xFF)

    def send_pex(self, connection):
        c = self.connections.get(connection)
        if c is not None:
            if self.config['ut_pex'] and not self.config['magnet'] and not self.paused:
                if c.ut_pex:
                    added = []
                    dropped = []
                    addedflags = []
                    curconnectiondns = []
                    for con in self.connections:
                        if con.remoteport is not None:
                            dns = (con.get_ip(), con.remoteport)
                            curconnectiondns.append(dns)
                            if con != connection and dns not in c.prevconnectiondns:
                                added.append(dns)
                                addedflag = 0
                                if con.cryptoallowed == 1:
                                    addedflag |= 1
                                if self.connections[con].download.have.complete():
                                    addedflag |= 2
                                addedflags.append(addedflag)
                    for dns in c.prevconnectiondns:
                        if dns not in curconnectiondns:
                            dropped.append(dns)
                    if added or dropped:
                        c.prevconnectiondns = curconnectiondns
                        pexdict = {}
                        pexdict['added'] = ''.join(map(self.compactPeerInfoFromDNS, added))
                        pexdict['added.f'] = ''.join(map(chr, addedflags))
                        pexdict['dropped'] = ''.join(map(self.compactPeerInfoFromDNS, dropped))
                        #print 'send_pex', pexdict
                        #print '...added :', added
                        #print '...added.f :', addedflags
                        #print '...dropped :', dropped
                        message = EXTENDED + chr(c.ut_pex) + bencode(pexdict)
                        c._send_message(message)
                        # Only rate measured message (because cyclical and sometimes lengthy)
                        self.pexupmeasure.update_rate(len(message))
                    c.utpexsendingrate = self.utpexsendingrate
            self.sched(self.send_pex, c.utpexsendingrate, [connection])

    def got_piece(self, i):
        for co in self.connections.values():
            co.send_have(i)

    def got_message(self, connection, message):
        c = self.connections[connection]
        t = message[0]
        if t != BITFIELD:
            c.download.gothavebitfield = True
        if t == BITFIELD and c.got_anything:
            connection.close()
            return
        if t != EXTENDED and t != PORT:
            c.got_anything = True
        if (t in [CHOKE, UNCHOKE, INTERESTED, NOT_INTERESTED] and 
            len(message) != 1):
            connection.close()
            return
        if t == CHOKE:
            c.download.got_choke()
        elif t == UNCHOKE:
            c.download.got_unchoke()
        elif t == INTERESTED:
            if not c.download.have.complete():
                c.upload.got_interested()
        elif t == NOT_INTERESTED:
            c.upload.got_not_interested()
        elif t == HAVE:
            if self.config['magnet']:
                return
            if len(message) != 5:
                connection.close()
                return
            i = toint(message[1:])
            if i >= self.numpieces:
                connection.close()
                return
            if c.download.got_have(i) and c.upload:
                c.upload.got_not_interested()
        elif t == BITFIELD:
            if self.config['magnet']:
                return
            try:
                b = Bitfield(self.numpieces, message[1:])
            except ValueError:
                connection.close()
                return
            if c.download.got_have_bitfield(b) and c.upload:
                c.upload.got_not_interested()
        elif t == REQUEST:
            if self.config['magnet']:
                return
            if len(message) != 13:
                connection.close()
                return
            i = toint(message[1:5])
            if i >= self.numpieces:
                connection.close()
                return
            c.got_request(i, toint(message[5:9]), toint(message[9:]))
        elif t == CANCEL:
            if self.config['magnet']:
                return
            if len(message) != 13:
                connection.close()
                return
            i = toint(message[1:5])
            if i >= self.numpieces:
                connection.close()
                return
            c.upload.got_cancel(i, toint(message[5:9]), toint(message[9:]))
        elif t == PIECE:
            if self.config['magnet']:
                return
            if len(message) <= 9:
                connection.close()
                return
            i = toint(message[1:5])
            if i >= self.numpieces:
                connection.close()
                return
            if c.download.got_piece(i, toint(message[5:9]), message[9:]):
                self.got_piece(i)
        elif t == PORT:
            if len(message) != 3:
                connection.close()
                return
            c.dht_port = ord(message[1]) << 8 | ord(message[2])
            if connection.uses_dht and self.dht:
                self.dht.sched(self.dht.addContact, 0, [c.get_ip(), c.dht_port])
        elif t == EXTENDED:
            try:
                extendeddict, extendeddictlength = bdecode(message[2:], sloppy = 1, length = True)
            except:
                c.ut_metadata = 0
                connection.close()
                return
            if message[1] == chr(0):
                if connection.got_extended:
                    return
                connection.got_extended = True
                if not 'm' in extendeddict:
                    c.ut_metadata = 0
                    c.ut_pex = 0
                    connection.close()
                    return
                c.got_extended(extendeddict)
            elif message[1] == chr(UT_METADATA):
                utmetadatamessageid = extendeddict.get('msg_type')
                piece = extendeddict.get('piece')
                if utmetadatamessageid is None or piece is None or type(piece) != int or piece < 0:
                    connection.close()
                    return
                if utmetadatamessageid == METADATAREQUEST:
                    c.got_metadatarequest(piece)
                elif utmetadatamessageid == METADATADATA:
                    totalsize = extendeddict.get('total_size')
                    if totalsize is None or len(message) > extendeddictlength + 16386:
                        connection.close()
                        return
                    c.got_metadatadata(piece, totalsize, message[extendeddictlength + 2:])
                elif utmetadatamessageid == METADATAREJECT:
                    c.got_metadatareject(piece)
                else:
                    connection.close()
            elif message[1] == chr(UT_PEX):
                if not self.config['ut_pex']:
                    return
                utpexadded = extendeddict.get('added')
                utpexaddedflags = extendeddict.get('added.f')
                # "dropped" key is sometimes missing from faulty clients
                if utpexadded is None or utpexaddedflags is None or len(utpexadded) % 6:
                    return
                # "added.f" may be empty from faulty clients
                if len(utpexaddedflags) == 0:
                    utpexaddedflags = '\x00' * (len(utpexadded) / 6)
                elif len(utpexaddedflags) * 6 != len(utpexadded):
                    return
                c.got_pex(utpexadded, utpexaddedflags)
            #else:
            #    connection.close()

        else:
            connection.close()

    def pause(self, flag):
        if flag:
            self.storePeers()
        if self.config['ut_pex']:
            self.switchPex(not flag)
        self.paused = flag

    def shutdown(self):
        self.connections = None
