# Written by Bram Cohen
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

from binascii import b2a_hex
from socket import error as socketerror
from urllib import quote
from BitTornado.BTcrypto import Crypto


DEBUG = False

MAX_INCOMPLETE = 8

protocol_name = 'BitTorrent protocol'


def toint(s):
    return long(b2a_hex(s), 16)

def tobinary16(i):
    return chr((i >> 8) & 0xFF) + chr(i & 0xFF)

hexchars = '0123456789ABCDEF'
hexmap = []
for i in xrange(256):
    hexmap.append(hexchars[(i&0xF0)/16]+hexchars[i&0x0F])

def tohex(s):
    r = []
    for c in s:
        r.append(hexmap[ord(c)])
    return ''.join(r)

def make_readable(s):
    if not s:
        return ''
    if quote(s).find('%') >= 0:
        return tohex(s)
    return '"'+s+'"'
   

# header, options, download id, my id, [length, message]

class Connection:
    def __init__(self, encoder, connection, id, option_pattern,
                 encrypted = None, options = None):
        self.encoder = encoder
        self.connection = connection
        self.connecter = encoder.connecter
        self.id = id
        self.option_pattern = option_pattern
        self.locally_initiated = (id != None)
        self.readable_id = make_readable(id)
        self.complete = False
        self.keepalive = lambda: None
        self.closed = False
        self.buffer = ''
        self.bufferlen = None
        self.log = None
        self.read = self._read
        self.write = self._write
        self.cryptmode = 0
        self.encrypter = None
        self.uses_dht = False
        self.uses_extended = False
        self.got_extended = False
        self.cryptoallowed = None
        self.remoteport = None

        if self.locally_initiated:
            self.encoder.incompletecounter += 1
            if encrypted:
                self.encrypted = True
                self.encrypter = Crypto(True)
                self.write(self.encrypter.pubkey + self.encrypter.padding())
            else:
                self.encrypted = False
                self.write(chr(len(protocol_name)) + protocol_name + 
                           self.option_pattern + self.encoder.download_id)
            self.next_len, self.next_func = 1 + len(protocol_name), self.read_header
        else:
            self.encrypted = None       # don't know yet
            self.next_len, self.next_func = 1 + len(protocol_name), self.read_header
        self.encoder.sched(self._auto_close, 15)

    def _log_start(self):   # only called with DEBUG = True
        self.log = open('peerlog.' + self.get_ip() + '.txt', 'a')
        self.log.write('connected - ')
        if self.locally_initiated:
            self.log.write('outgoing\n')
        else:
            self.log.write('incoming\n')
        self._logwritefunc = self.write
        self.write = self._log_write

    def _log_write(self, s):
        self.log.write('w:' + b2a_hex(s) + '\n')
        self._logwritefunc(s)

    def get_ip(self, real = False):
        return self.connection.get_ip(real)

    def get_port(self, real = False):
        return self.connection.get_port(real)

    def set_port(self, port):
        self.connection.set_port(port)

    def get_dns(self, real = False):
        return self.connection.get_dns(real)

    def get_id(self):
        return self.id

    def get_readable_id(self):
        return self.readable_id

    def is_locally_initiated(self):
        return self.locally_initiated

    def is_encrypted(self):
        return bool(self.encrypted)

    def is_flushed(self):
        return self.connection.is_flushed()

    def _read_header(self, s):
        if s == chr(len(protocol_name)) + protocol_name:
            return 8, self.read_options
        return None

    def read_header(self, s):
        #print 'read_header'
        if self._read_header(s):
            if self.encrypted or self.encoder.config['crypto_stealth']:
                return None
            return 8, self.read_options
        if self.locally_initiated and not self.encrypted:
            return None
        elif not self.encoder.config['crypto_allowed']:
            return None
        if not self.encrypted:
            self.encrypted = True
            self.encrypter = Crypto(self.locally_initiated)
        self._write_buffer(s)
        return self.encrypter.keylength, self.read_crypto_header

    ################## ENCRYPTION SUPPORT ######################

    def _start_crypto(self):
        self.encrypter.setrawaccess(self._read, self._write)
        self.write = self.encrypter.write
        self.read = self.encrypter.read
        if self.buffer:
            self.buffer = self.encrypter.decrypt(self.buffer)

    def _end_crypto(self):
        self.read = self._read
        self.write = self._write
        self.encrypter = None

    def read_crypto_header(self, s):
        self.encrypter.received_key(s)
        self.encrypter.set_skey(self.encoder.download_id)
        if self.locally_initiated:
            if self.encoder.config['crypto_only']:
                cryptmode = '\x00\x00\x00\x02'    # full stream encryption
            else:
                cryptmode = '\x00\x00\x00\x03'    # header or full stream
            padc = self.encrypter.padding()
            self.write(self.encrypter.block3a
                       + self.encrypter.block3b
                       + self.encrypter.encrypt(
                       ('\x00' * 8)               # VC
                       + cryptmode                # acceptable crypto modes
                       + tobinary16(len(padc))
                       + padc                     # PadC
                       + '\x00\x00'))             # no initial payload data
            self._max_search = 520
            return 1, self.read_crypto_block4a
        self.write(self.encrypter.pubkey + self.encrypter.padding())
        self._max_search = 520
        return 0, self.read_crypto_block3a

    def _search_for_pattern(self, s, pat):
        p = s.find(pat)
        if p < 0:
            if len(s) >= len(pat):
                self._max_search -= len(s) + 1 - len(pat)
            if self._max_search < 0:
                self.close()
                return False
            self._write_buffer(s[1 - len(pat):])
            return False
        self._write_buffer(s[p + len(pat):])
        return True

    ### INCOMING CONNECTION ###
    def read_crypto_block3a(self, s):
        if not self._search_for_pattern(s, self.encrypter.block3a):
            return -1, self.read_crypto_block3a     # wait for more data
        return len(self.encrypter.block3b), self.read_crypto_block3b

    def read_crypto_block3b(self, s):
        if s != self.encrypter.block3b:
            return None
        self.encoder.connecter.external_connection_made += 1
        self._start_crypto()
        return 14, self.read_crypto_block3c

    def read_crypto_block3c(self, s):
        if s[:8] != ('\x00' * 8):           # check VC
            return None
        self.cryptmode = toint(s[8:12]) % 4
        if self.cryptmode == 0:
            return None                     # no encryption selected
        if ( self.cryptmode == 1            # only header encryption
             and self.encoder.config['crypto_only'] ):
            return None
        padlen = (ord(s[12]) << 8) + ord(s[13])
        if padlen > 512:
            return None
        return padlen + 2, self.read_crypto_pad3

    def read_crypto_pad3(self, s):
        s = s[-2:]
        ialen = (ord(s[0]) << 8) + ord(s[1])
        if ialen > 65535:
            return None
        if self.cryptmode == 1:
            cryptmode = '\x00\x00\x00\x01'    # header only encryption
        else:
            cryptmode = '\x00\x00\x00\x02'    # full stream encryption
        padd = self.encrypter.padding()
        self.write(('\x00' * 8)            # VC
                   + cryptmode             # encryption mode
                   + tobinary16(len(padd))
                   + padd)                 # PadD
        if ialen:
            return ialen, self.read_crypto_ia
        return self.read_crypto_block3done()

    def read_crypto_ia(self, s):
        if DEBUG:
            self._log_start()
            self.log.write('r:' + b2a_hex(s) + '(ia)\n')
            if self.buffer:
                self.log.write('r:' + b2a_hex(self.buffer) + '(buffer)\n')
        return self.read_crypto_block3done(s)

    def read_crypto_block3done(self, ia = ''):
        if DEBUG:
            if not self.log:
                self._log_start()
        if self.cryptmode == 1:     # only handshake encryption
            assert not self.buffer  # oops; check for exceptions to this
            self._end_crypto()
        if ia:
            self._write_buffer(ia)
        return 1 + len(protocol_name), self.read_encrypted_header

    ### OUTGOING CONNECTION ###
    def read_crypto_block4a(self, s):
        if not self._search_for_pattern(s, self.encrypter.VC_pattern()):
            return -1, self.read_crypto_block4a     # wait for more data
        self._start_crypto()
        return 6, self.read_crypto_block4b

    def read_crypto_block4b(self, s):
        self.cryptmode = toint(s[:4]) % 4
        if self.cryptmode == 1:             # only header encryption
            if self.encoder.config['crypto_only']:
                return None
        elif self.cryptmode != 2:
            return None                     # unknown encryption
        padlen = (ord(s[4]) << 8) + ord(s[5])
        if padlen > 512:
            return None
        if padlen:
            return padlen, self.read_crypto_pad4
        return self.read_crypto_block4done()

    def read_crypto_pad4(self, s):
        # discard data
        return self.read_crypto_block4done()

    def read_crypto_block4done(self):
        if DEBUG:
            self._log_start()
        if self.cryptmode == 1:     # only handshake encryption
            if not self.buffer:  # oops; check for exceptions to this
                return None
            self._end_crypto()
        self.write(chr(len(protocol_name)) + protocol_name + 
                   self.option_pattern + self.encoder.download_id)
        return 1 + len(protocol_name), self.read_encrypted_header

    ### START PROTOCOL OVER ENCRYPTED CONNECTION ###
    def read_encrypted_header(self, s):
        return self._read_header(s)

    ################################################

    def read_options(self, s):
        self.options = s
        if ord(self.options[7]) & 1:
            self.uses_dht = True
        if ord(self.options[5]) & 16:
            self.uses_extended = True
        return 20, self.read_download_id

    def read_download_id(self, s):
        if (s != self.encoder.download_id
            or not self.encoder.check_ip(ip = self.get_ip())):
            return None
        if not self.locally_initiated:
            if not self.encrypted:
                self.encoder.connecter.external_connection_made += 1
            self.write(chr(len(protocol_name)) + protocol_name + 
                       self.option_pattern + self.encoder.download_id + self.encoder.my_id)
        return 20, self.read_peer_id

    def read_peer_id(self, s):
        if not self.encrypted and self.encoder.config['crypto_only']:
            return None     # allows older trackers to ping,
                            # but won't proceed w/ connections
        if not self.id:
            self.id = s
            self.readable_id = make_readable(s)
        else:
            if s != self.id:
                return None
        self.complete = self.encoder.got_id(self)
        if not self.complete:
            return None
        if self.locally_initiated:
            self.write(self.encoder.my_id)
            self.encoder.incompletecounter -= 1
        self._switch_to_read2()
        c = self.encoder.connecter.connection_made(self)
        self.keepalive = c.send_keepalive
        return 4, self.read_len

    def read_len(self, s):
        l = toint(s)
        if l > self.encoder.max_len:
            return None
        return l, self.read_message

    def read_message(self, s):
        if s != '':
            self.connecter.got_message(self, s)
        return 4, self.read_len

    def read_dead(self, s):
        return None

    def _auto_close(self):
        if not self.complete:
            self.close()

    def close(self, shutdown = False, trynext = True):
        #print 'Closing connection'
        if not self.closed:
            self.connection.close()
            self.sever(shutdown, trynext)

    def sever(self, shutdown = False, trynext = True):
        if self.log:
            self.log.write('closed\n')
            self.log.close()
        self.closed = True
        del self.encoder.connections[self.connection]
        if self.complete:
            self.connecter.connection_lost(self, shutdown)
        elif self.locally_initiated:
            self.encoder.incompletecounter -= 1
        if self.encoder.con is self:
            self.encoder.con = None
        self.connection = None
        if not shutdown and trynext and self.locally_initiated and self.encoder.config['magnet']:
            self.encoder.nextConnection()

    def send_message_raw(self, message):
        self.write(message)

    def _write(self, message):
        if not self.closed:
            self.connection.write(message)

    def data_came_in(self, connection, s):
        self.read(s)

    def _write_buffer(self, s):
        self.buffer = s + self.buffer

    def _read(self, s):
        if self.log:
            self.log.write('r:' + b2a_hex(s) + '\n')
        self.encoder.measurefunc(len(s))
        self.buffer += s
        while True:
            if self.closed:
                return
            # self.next_len = # of characters function expects
            # or 0 = all characters in the buffer
            # or -1 = wait for next read, then all characters in the buffer
            # not compatible w/ keepalives, switch out after all negotiation complete
            if self.next_len <= 0:
                m = self.buffer
                self.buffer = ''
            elif len(self.buffer) >= self.next_len:
                m = self.buffer[:self.next_len]
                self.buffer = self.buffer[self.next_len:]
            else:
                return
            try:
                x = self.next_func(m)
            except:
                self.next_len, self.next_func = 1, self.read_dead
                raise
            if x is None:
                self.close()
                return
            self.next_len, self.next_func = x
            if self.next_len < 0:  # already checked buffer
                return             # wait for additional data
            if self.bufferlen is not None:
                self._read2('')
                return

    def _switch_to_read2(self):
        self._write_buffer = None
        if self.encrypter:
            self.encrypter.setrawaccess(self._read2, self._write)
        else:
            self.read = self._read2
        self.bufferlen = len(self.buffer)
        self.buffer = [self.buffer]

    def _read2(self, s):    # more efficient, requires buffer['',''] & bufferlen
        if self.log:
            self.log.write('r:' + b2a_hex(s) + '\n')
        self.encoder.measurefunc(len(s))
        while True:
            if self.closed:
                return
            p = self.next_len - self.bufferlen
            if self.next_len == 0:
                m = ''
            elif s:
                if p > len(s):
                    self.buffer.append(s)
                    self.bufferlen += len(s)
                    return
                self.bufferlen = len(s) - p
                self.buffer.append(s[:p])
                m = ''.join(self.buffer)
                if p == len(s):
                    self.buffer = []
                else:
                    self.buffer = [s[p:]]
                s = ''
            elif p <= 0:
                # assert len(self.buffer) == 1
                s = self.buffer[0]
                self.bufferlen = len(s) - self.next_len
                m = s[:self.next_len]
                if p == 0:
                    self.buffer = []
                else:
                    self.buffer = [s[self.next_len:]]
                s = ''
            else:
                return
            try:
                x = self.next_func(m)
            except:
                self.next_len, self.next_func = 1, self.read_dead
                raise
            if x is None:
                self.close()
                return
            self.next_len, self.next_func = x
            if self.next_len < 0:  # already checked buffer
                return             # wait for additional data

    def connection_flushed(self, connection):
        if self.complete:
            self.connecter.connection_flushed(self)

    def connection_lost(self, connection):
        if self.encoder.connections.has_key(connection):
            #print 'Connection lost'
            self.sever()


class _dummy_banlist:
    def includes(self, x):
        return False


class Encoder:
    def __init__(self, connecter, raw_server, my_id,
                 dht, sched, download_id, measurefunc, config,
                 bans = _dummy_banlist()):
        self.config = config
        self.raw_server = raw_server
        self.connecter = connecter
        self.my_id = my_id
        self.max_len = self.config['max_message_length']
        self.dht = dht
        self.sched = sched
        self.keepalive_delay = self.config['keepalive_interval']
        self.download_id = download_id
        self.measurefunc = measurefunc
        self.connections = {}
        self.banned = {}
        self.external_bans = bans
        self.to_connect = []
        self.connected_for_metadata = []
        self.no_metadata_support = []
        self.paused = False
        if self.config['max_connections'] == 0:
            self.max_connections = 2 ** 30
        else:
            self.max_connections = self.config['max_connections']
        sched(self.send_keepalives, self.keepalive_delay)
        if self.dht:
            # extended messages + dht
            self.option_pattern = chr(0) * 5 + chr(16) + chr(0) + chr(1)
        else:
            # extended messages
            self.option_pattern = chr(0) * 5 + chr(16) + chr(0) * 2
        self.incompletecounter = 0
        self.downloadingmetadata = False
        # Current meatadata request (-2 : none, -1 : extended, >=0 : piece request)
        self.curmetadatarequest = -2
        self.metadatatimeout = self.config['metadata_timeout']
        self.con = None

    def send_keepalives(self):
        self.sched(self.send_keepalives, self.keepalive_delay)
        if self.paused:
            return
        for c in self.connections.values():
            c.keepalive()

    def initPeers(self):
        if self.connecter.peers:
            self.sched(self.start_connections, 0, [[(p, 0, None, None) for p in self.connecter.peers]])

    def start_connections(self, list):
        #print 'start_connections', list, self.downloadingmetadata
        if not self.to_connect and not self.downloadingmetadata:
            self.sched(self._start_connection_from_queue)
        self.to_connect.extend(list)

    def _start_connection_from_queue(self):
        #print '_start_connection_from_queue'
        if not self.to_connect:
            return
        if self.connecter.external_connection_made:
            max_initiate = self.config['max_initiate']
        else:
            max_initiate = int(self.config['max_initiate'] * 1.5)
        cons = len(self.connections)
        if cons >= self.max_connections or cons >= max_initiate:
            #print 'case 1 :', cons, self.max_connections, max_initiate
            delay = 60
        elif self.paused or self.incompletecounter >= MAX_INCOMPLETE:
            #print 'case 2 :', self.paused, self.incompletecounter >= MAX_INCOMPLETE
            delay = 1
        else:
            #print 'case 3'
            delay = 0
            self.curpeer = self.to_connect.pop(0)
            self.start_connection(self.curpeer)
        if self.to_connect and not self.downloadingmetadata:
            self.sched(self._start_connection_from_queue, delay)

    def start_connection(self, curpeer):
        #print 'start_connection'
        dns, id, encrypted, cryptoallowed = curpeer
        if (self.paused
            or len(self.connections) >= self.max_connections
            or id == self.my_id
            or not self.check_ip(ip = dns[0])):
            return True
        if self.config['crypto_only']:
            if encrypted == 0 or cryptoallowed == 0:
                return True
            encrypted = True
        if self.config['magnet']:
            for d, i, e, c in self.no_metadata_support:
                if id and i == id or d == dns:
                    #print 'Skipping already contacted peer with no metadata support'
                    return True
        else:
            sec = self.config['security']
            for v in self.connections.values():
                if v is None:
                    continue
                condns = v.get_dns(True)
                if id and v.id == id or condns == dns:
                    if v.remoteport is None:
                        v.remoteport = dns[1]
                    if v.cryptoallowed is None:
                        v.cryptoallowed = cryptoallowed
                    return True
                if sec and condns[0] != 'unknown' and condns[0] == dns[0]:
                    return True
        try:
            #print 'Trying to create connection'
            c = self.raw_server.start_connection(dns)
            #print 'Creating connection'
            self.con = Connection(self, c, id, self.option_pattern, encrypted = encrypted)
            self.connections[c] = self.con
            c.set_handler(self.con)
        except socketerror:
            return False

        if self.config['magnet']:
            self.downloadingmetadata = True
            for d, i, e, c in self.connected_for_metadata:
                if id and i == id or d == dns:
                    break
            else:
                self.connected_for_metadata.append(curpeer)

        return True

    def _start_connection(self, dns, id, encrypted = None):
        self.sched(self.start_connection, 0, [(dns, id, encrypted)])

    def startMetadataTimeOut(self, request):
        self.curmetadatarequest = request
        self.sched(self.onMetadataTimeOut, self.metadatatimeout, [self.con, request])

    def resetMetadataTimeOut(self):
        self.curmetadatarequest = -2

    def onMetadataTimeOut(self, con, request):
        if con is self.con and request == self.curmetadatarequest:
            self.con.close()

    def nextConnection(self):
        #print 'Next connection'
        self.downloadingmetadata = False
        self.sched(self._start_connection_from_queue)

    def onNoMetadata(self):
        self.sched(self._onNoMetadata)

    def _onNoMetadata(self):
        self.connecter.resetMetadata()
        self.nextConnection()

    def onMetadataOK(self):
        self.sched(self._onMetadataOK)

    def _onMetadataOK(self):
        if self.connected_for_metadata:
            self.to_connect.insert(0, self.connected_for_metadata[-1])
            self.to_connect.extend(self.connected_for_metadata[:-1])
            self.connected_for_metadata[:] = []
        self.no_metadata_support[:] = []
        self.config['magnet'] = 0
        self.nextConnection()

    def storePeerWithNoMetadataSupport(self):
        #print 'Storing peer with no metadata support'
        self.no_metadata_support.append(self.curpeer)

    def check_ip(self, connection = None, ip = None):
        if not ip:
            ip = connection.get_ip(True)
        if self.config['security'] and self.banned.has_key(ip):
            return False
        if self.external_bans.includes(ip):
            return False
        return True

    def got_id(self, connection):
        if connection.id == self.my_id:
            self.connecter.external_connection_made -= 1
            return False
        ip = connection.get_ip(True)
        sec = self.config['security']
        for v in self.connections.values():
            if v is not connection:
                if connection.id == v.id:
                    # There will be at most 1 duplicate id
                    if ip == v.get_ip(True):
                        if connection.remoteport is None:
                            connection.remoteport = v.remoteport
                        if connection.cryptoallowed is None:
                            connection.cryptoallowed = v.cryptoallowed
                        v.close
                    return False
                elif sec and ip != 'unknown' and ip == v.get_ip(True):
                    v.close
        return True

    def external_connection_made(self, connection):
        if self.paused or self.config['magnet'] or len(self.connections) >= self.max_connections \
           or not self.check_ip(connection=connection):
            connection.close()
            return False
        con = Connection(self, connection, None, self.option_pattern)
        self.connections[connection] = con
        connection.set_handler(con)
        return True

    def ban(self, ip):
        self.banned[ip] = 1

    def pause(self, flag):
        if not flag:
            self.initPeers()
        self.paused = flag

    def shutdown(self):
        for c in self.connections.values():
            c.close(shutdown = True)
        self.connections = None
